@@ -11,18 +11,12 @@
 CryptData::CryptData()
 {
   Method=CRYPT_NONE;
-  memset(KDFCache,0,sizeof(KDFCache));
-  KDFCachePos=0;
+  KDF3CachePos=0;
+  KDF5CachePos=0;
   memset(CRCTab,0,sizeof(CRCTab));
 }
 
 
-CryptData::~CryptData()
-{
-  cleandata(KDFCache,sizeof(KDFCache));
-}
-
-
 
 
 void CryptData::DecryptBlock(byte *Buf,size_t Size)
@@ -49,20 +43,32 @@ void CryptData::DecryptBlock(byte *Buf,size_t Size)
 }
 
 
-void CryptData::SetCryptKeys(bool Encrypt,CRYPT_METHOD Method,
+bool CryptData::SetCryptKeys(bool Encrypt,CRYPT_METHOD Method,
      SecPassword *Password,const byte *Salt,
      const byte *InitV,uint Lg2Cnt,byte *HashKey,byte *PswCheck)
 {
-  if (!Password->IsSet() || Method==CRYPT_NONE)
-    return;
+  if (Method==CRYPT_NONE || !Password->IsSet())
+    return false;
 
   CryptData::Method=Method;
 
   wchar PwdW[MAXPASSWORD];
   Password->Get(PwdW,ASIZE(PwdW));
+
+  // Display this warning only when encrypting. Users complained that
+  // it is distracting when decrypting. It still can be useful when encrypting,
+  // so users do not waste time to excessively long passwords.
+  if (Encrypt && wcslen(PwdW)>=MAXPASSWORD_RAR)
+    uiMsg(UIERROR_TRUNCPSW,MAXPASSWORD_RAR-1);
+
+  PwdW[Min(MAXPASSWORD_RAR,MAXPASSWORD)-1]=0; // For compatibility with existing archives.
+
   char PwdA[MAXPASSWORD];
   WideToChar(PwdW,PwdA,ASIZE(PwdA));
+  PwdA[Min(MAXPASSWORD_RAR,MAXPASSWORD)-1]=0; // For compatibility with existing archives.
 
+  bool Success=true;
+  
   switch(Method)
   {
 #ifndef SFX_MODULE
@@ -78,13 +84,29 @@ void CryptData::SetCryptKeys(bool Encrypt,CRYPT_METHOD Method,
 #endif
     case CRYPT_RAR30:
       SetKey30(Encrypt,Password,PwdW,Salt);
-      return;
+      break;
     case CRYPT_RAR50:
-      SetKey50(Encrypt,Password,PwdW,Salt,InitV,Lg2Cnt,HashKey,PswCheck);
-      return;
+      Success=SetKey50(Encrypt,Password,PwdW,Salt,InitV,Lg2Cnt,HashKey,PswCheck);
+      break;
   }
   cleandata(PwdA,sizeof(PwdA));
   cleandata(PwdW,sizeof(PwdW));
+  return Success;
+}
+
+
+// Use the current system time to additionally randomize data.
+static void TimeRandomize(byte *RndBuf,size_t BufSize)
+{
+  static uint Count=0;
+  RarTime CurTime;
+  CurTime.SetCurrentTime();
+  uint64 Random=CurTime.GetWin()+clock();
+  for (size_t I=0;I<BufSize;I++)
+  {
+    byte RndByte = byte (Random >> ( (I & 7) * 8 ));
+    RndBuf[I]=byte( (RndByte ^ I) + Count++);
+  }
 }
 
 
@@ -96,10 +118,10 @@ void GetRnd(byte *RndBuf,size_t BufSize)
   bool Success=false;
 #if defined(_WIN_ALL)
   HCRYPTPROV hProvider = 0;
-	if (CryptAcquireContext(&hProvider, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT))
+  if (CryptAcquireContext(&hProvider, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT))
   {
-	  Success=CryptGenRandom(hProvider, (DWORD)BufSize, RndBuf) == TRUE;
-	  CryptReleaseContext(hProvider, 0);
+    Success=CryptGenRandom(hProvider, (DWORD)BufSize, RndBuf) != FALSE;
+    CryptReleaseContext(hProvider, 0);
   }
 #elif defined(_UNIX)
   FILE *rndf = fopen("/dev/urandom", "r");
@@ -111,15 +133,5 @@ void GetRnd(byte *RndBuf,size_t BufSize)
 #endif
   // We use this code only as the last resort if code above failed.
   if (!Success)
-  {
-    static uint Count=0;
-    RarTime CurTime;
-    CurTime.SetCurrentTime();
-    uint64 Random=CurTime.GetRaw()+clock();
-    for (size_t I=0;I<BufSize;I++)
-    {
-      byte RndByte = byte (Random >> ( (I & 7) * 8 ));
-      RndBuf[I]=byte( (RndByte ^ I) + Count++);
-    }
-  }
+    TimeRandomize(RndBuf,BufSize);
 }
